import numbers
from six import string_types
from pypif.obj.common.id import Id
from pypif.obj.common.process_step import ProcessStep
from pypif.obj.common.property import Property
from pypif.obj.common.quantity import Quantity
from pypif.obj.common.rcl import Rcl
from pypif.obj.common.source import Source


class System(Rcl):
    """
    Base representation for all systems.
    """

    def __init__(self, uid=None, names=None, ids=None, source=None, quantity=None, properties=None, preparation=None,
                 sub_systems=None, references=None, contacts=None, licenses=None, tags=None, **kwargs):
        """
        Constructor.

        :param uid: String with the permanent ID for this record.
        :param names: List of strings with common names of the system.
        :param ids: List of dictionaries, strings, numbers, or :class:`.Id` objects that identify the system.
        :param source: Dictionary, string, or :class:`.Source` object with the source of the system.
        :param quantity: Dictionary or :class:`.Quantity` object with the quantity of the system.
        :param properties: List of dictionaries or :class:`.Property` objects with properties of the system.
        :param preparation: List of dictionaries or :class:`.ProcessStep` objects with the preparation
                information of the system.
        :param sub_systems: List of dictionaries or :class:`.System` objects with the subsystems of the system.
        :param references: List of dictionaries or :class:`.Reference` objects where information about the system
                is published.
        :param contacts: List of dictionaries, strings, or :class:`.Person` objects with people to contact for
                information about the system.
        :param licenses: List of dictionaries, strings, or :class:`.License` objects with licensing information
                for data about the system.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(System, self).__init__(references=references, contacts=contacts, licenses=licenses, tags=tags, **kwargs)
        self._uid = None
        self.uid = uid
        self._names = None
        self.names = names
        self._ids = None
        self.ids = ids
        self._source = None
        self.source = source
        self._quantity = None
        self.quantity = quantity
        self._properties = None
        self.properties = properties
        self._preparation = None
        self.preparation = preparation
        self._sub_systems = None
        self.sub_systems = sub_systems
        self.category = kwargs['category'] if 'category' in kwargs else 'system'

    @property
    def uid(self):
        return self._uid

    @uid.setter
    def uid(self, uid):
        self._validate_type('uid', uid, string_types)
        self._uid = uid

    @uid.deleter
    def uid(self):
        self._uid = None

    @property
    def names(self):
        return self._names

    @names.setter
    def names(self, names):
        self._validate_list_type('names', names, string_types)
        self._names = names

    @names.deleter
    def names(self):
        self._names = None

    @property
    def ids(self):
        return self._ids

    @ids.setter
    def ids(self, ids):
        self._validate_list_type('ids', ids, dict, string_types, numbers.Number, Id)
        self._ids = self._get_object(Id, ids)

    @ids.deleter
    def ids(self):
        self._ids = None

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        self._validate_type('source', source, dict, string_types, Source)
        self._source = self._get_object(Source, source)

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, quantity):
        self._validate_type('quantity', quantity, dict, Quantity)
        self._quantity = self._get_object(Quantity, quantity)

    @quantity.deleter
    def quantity(self):
        self._quantity = None

    @source.deleter
    def source(self):
        self._source = None

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, properties):
        self._validate_list_type('properties', properties, dict, Property)
        self._properties = self._get_object(Property, properties)

    @properties.deleter
    def properties(self):
        self._properties = None

    @property
    def preparation(self):
        return self._preparation

    @preparation.setter
    def preparation(self, preparation):
        self._validate_list_type('preparation', preparation, dict, ProcessStep)
        self._preparation = self._get_object(ProcessStep, preparation)

    @preparation.deleter
    def preparation(self):
        self._preparation = None

    @property
    def sub_systems(self):
        return self._sub_systems

    @sub_systems.setter
    def sub_systems(self, sub_systems):
        self._validate_list_type('sub_systems', sub_systems, dict, System)
        self._sub_systems = self._get_object(System, sub_systems)

    @sub_systems.deleter
    def sub_systems(self):
        self._sub_systems = None



from six import string_types
from pypif.obj.common.display_item import DisplayItem
from pypif.obj.common.name import Name
from pypif.obj.common.pages import Pages
from pypif.obj.common.pio import Pio


class Reference(Pio):
    """
    Information about a referenced publication.
    """

    def __init__(self, doi=None, isbn=None, issn=None, url=None, title=None, publisher=None, journal=None, volume=None,
                 issue=None, year=None, figure=None, table=None, pages=None, authors=None, editors=None,
                 affiliations=None, acknowledgements=None, references=None, tags=None, **kwargs):
        """
        Constructor.

        :param doi: String with DOI of the published work
        :param isbn: String with ISBN of the published work
        :param issn: String with ISSN of the published work
        :param url: String with URL to the published work
        :param title: String with title of the published work.
        :param publisher: String with publisher of the work.
        :param journal: String with the journal in which the work was published.
        :param volume: String with the volume in which the work was published.
        :param issue: String with the issue in which the work was published.
        :param year: String with the year in which the work was published.
        :param figure: Dictionary or :class:`.DisplayItem` object with the figure to reference.
        :param table: Dictionary or :class:`.DisplayItem` object with the table to reference.
        :param pages: String, integer, dictionary, or :class:`.Pages` object with the starting and ending pages for
                the published work.
        :param authors: List of strings, dictionaries, or :class:`.Name` objects with information about the authors.
        :param editors: List of strings, dictionaries, or :class:`.Name` objects with information about the editors.
        :param affiliations: List of strings with affiliations.
        :param acknowledgements: List of strings with acknowledgements.
        :param references: List of dictionaries or :class:`.Reference` objects with works cited by this published work.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Reference, self).__init__(tags=tags, **kwargs)
        self._doi = None
        self.doi = doi
        self._isbn = None
        self.isbn = isbn
        self._issn = None
        self.issn = issn
        self._url = None
        self.url = url
        self._title = None
        self.title = title
        self._publisher = None
        self.publisher = publisher
        self._journal = None
        self.journal = journal
        self._volume = None
        self.volume = volume
        self._issue = None
        self.issue = issue
        self._year = None
        self.year = year
        self._figure = None
        self.figure = figure
        self._table = None
        self.table = table
        self._pages = None
        self.pages = pages
        self._authors = None
        self.authors = authors
        self._editors = None
        self.editors = editors
        self._affiliations = None
        self.affiliations = affiliations
        self._acknowledgements = None
        self.acknowledgements = acknowledgements
        self._references = None
        self.references = references

    @property
    def doi(self):
        return self._doi

    @doi.setter
    def doi(self, doi):
        self._validate_type('doi', doi, string_types)
        self._doi = doi

    @doi.deleter
    def doi(self):
        self._doi = None

    @property
    def isbn(self):
        return self._isbn

    @isbn.setter
    def isbn(self, isbn):
        self._validate_type('isbn', isbn, string_types)
        self._isbn = isbn

    @isbn.deleter
    def isbn(self):
        self._isbn = None

    @property
    def issn(self):
        return self.issn

    @issn.setter
    def issn(self, issn):
        self._validate_type('issn', issn, string_types)
        self._issn = issn

    @issn.deleter
    def issn(self):
        self._issn = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._validate_type('url', url, string_types)
        self._url = url

    @url.deleter
    def url(self):
        self._url = None

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self._validate_type('title', title, string_types)
        self._title = title

    @title.deleter
    def title(self):
        self._title = None

    @property
    def publisher(self):
        return self._publisher

    @publisher.setter
    def publisher(self, publisher):
        self._validate_type('publisher', publisher, string_types)
        self._publisher = publisher

    @publisher.deleter
    def publisher(self):
        self._publisher = None

    @property
    def journal(self):
        return self._journal

    @journal.setter
    def journal(self, journal):
        self._validate_type('journal', journal, string_types)
        self._journal = journal

    @journal.deleter
    def journal(self):
        self._journal = None

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, volume):
        self._validate_type('volume', volume, string_types)
        self._volume = volume

    @volume.deleter
    def volume(self):
        self._volume = None

    @property
    def issue(self):
        return self._issue

    @issue.setter
    def issue(self, issue):
        self._validate_type('issue', issue, string_types)
        self._issue = issue

    @issue.deleter
    def issue(self):
        self._issue = None

    @property
    def year(self):
        return self._year

    @year.setter
    def year(self, year):
        self._validate_type('year', year, string_types)
        self._year = year

    @year.deleter
    def year(self):
        self._year = None

    @property
    def figure(self):
        return self._figure

    @figure.setter
    def figure(self, figure):
        self._validate_type('figure', figure, dict, DisplayItem)
        self._figure = self._get_object(DisplayItem, figure)

    @figure.deleter
    def figure(self):
        self._figure = None

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, table):
        self._validate_type('table', table, dict, DisplayItem)
        self._table = self._get_object(DisplayItem, table)

    @table.deleter
    def table(self):
        self._table = None

    @property
    def pages(self):
        return self._pages

    @pages.setter
    def pages(self, pages):
        self._validate_type('pages', pages, string_types, int, dict, Pages)
        self._pages = self._get_object(Pages, pages)

    @pages.deleter
    def pages(self):
        self._pages = None

    @property
    def authors(self):
        return self._authors

    @authors.setter
    def authors(self, authors):
        self._validate_list_type('authors', authors, string_types, dict, Name)
        self._authors = self._get_object(Name, authors)

    @authors.deleter
    def authors(self):
        self._authors = None

    @property
    def editors(self):
        return self._editors

    @editors.setter
    def editors(self, editors):
        self._validate_list_type('editors', editors, string_types, dict, Name)
        self._editors = self._get_object(Name, editors)

    @editors.deleter
    def editors(self):
        self._editors = None

    @property
    def affiliations(self):
        return self._affiliations

    @affiliations.setter
    def affiliations(self, affiliations):
        self._validate_list_type('affiliations', affiliations, string_types)
        self._affiliations = affiliations

    @affiliations.deleter
    def affiliations(self):
        self._affiliations = None

    @property
    def acknowledgements(self):
        return self._acknowledgements

    @acknowledgements.setter
    def acknowledgements(self, acknowledgements):
        self._validate_list_type('acknowledgements', acknowledgements, string_types)
        self._acknowledgements = acknowledgements

    @acknowledgements.deleter
    def acknowledgements(self):
        self._acknowledgements = None

    @property
    def references(self):
        return self._references

    @references.setter
    def references(self, references):
        self._validate_list_type('references', references, dict, Reference)
        self._references = self._get_object(Reference, references)



from six import string_types
from pypif.obj.common.pio import Pio


class License(Pio):
    """
    Information about a license that applies to some item.
    """

    def __init__(self, name=None, description=None, url=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the license.
        :param description: String with the description of the license.
        :param url: String with the URL to the license.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(License, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._description = None
        self.description = description
        self._url = None
        self.url = url

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, string_types)
        self._name = name

    @name.deleter
    def name(self):
        self._name = None

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._validate_type('description', description, string_types)
        self._description = description

    @description.deleter
    def description(self):
        self._description = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._validate_type('url', url, string_types)
        self._url = url

    @url.deleter
    def url(self):
        self._url = None



import numbers
from six import string_types
from pypif.obj.common.pio import Pio


class Id(Pio):
    """
    Information about a generic identifier.
    """

    def __init__(self, name=None, value=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the identifier.
        :param value: String or number with the value of the identifier.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Id, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._value = None
        self.value = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, string_types)
        self._name = name

    @name.deleter
    def name(self):
        self._name = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._validate_list_type('value', value, string_types, numbers.Number)
        self._value = value

    @value.deleter
    def value(self):
        self._value = None



from pypif.obj.common.pio import Pio
from pypif.obj.common.scalar import Scalar


class Quantity(Pio):
    """
    Information about the quantity of a system.
    """

    def __init__(self, actual_mass_percent=None, actual_volume_percent=None, actual_number_percent=None,
                 ideal_mass_percent=None, ideal_volume_percent=None, ideal_number_percent=None, tags=None, **kwargs):
        """
        Constructor.
        
        :param actual_mass_percent: Dictionary or :class:`.Scalar` object with the actual percent of the total mass 
                made up by this system.
        :param actual_volume_percent: Dictionary or :class:`.Scalar` object with the actual percent of the total 
                volume made up by this system.
        :param actual_number_percent: Dictionary or :class:`.Scalar` object with the actual percent of the total 
                numeric quantity made up by this system.
        :param ideal_mass_percent: Dictionary or :class:`.Scalar` object with the ideal percent of the total mass 
                made up by this system.
        :param ideal_volume_percent: Dictionary or :class:`.Scalar` object with the ideal percent of the total 
                volume made up by this system.
        :param ideal_number_percent: Dictionary or :class:`.Scalar` object with the ideal percent of the total 
                numeric quantity made up by this system.
        :param tags: List of tags that apply to the quantity.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Quantity, self).__init__(tags=tags, **kwargs)
        self._actual_mass_percent = None
        self.actual_mass_percent = actual_mass_percent
        self._actual_volume_percent = None
        self.actual_volume_percent = actual_volume_percent
        self._actual_number_percent = None
        self.actual_number_percent = actual_number_percent
        self._ideal_mass_percent = None
        self.ideal_mass_percent = ideal_mass_percent
        self._ideal_volume_percent = None
        self.ideal_volume_percent = ideal_volume_percent
        self._ideal_number_percent = None
        self.ideal_number_percent = ideal_number_percent

    @property
    def actual_mass_percent(self):
        return self._actual_mass_percent
    
    @actual_mass_percent.setter
    def actual_mass_percent(self, actual_mass_percent):
        self._validate_type('actual_mass_percent', actual_mass_percent, dict, Scalar)
        self._actual_mass_percent = self._get_object(Scalar, actual_mass_percent)
    
    @actual_mass_percent.deleter
    def actual_mass_percent(self):
        self._actual_mass_percent = None

    @property
    def actual_volume_percent(self):
        return self._actual_volume_percent

    @actual_volume_percent.setter
    def actual_volume_percent(self, actual_volume_percent):
        self._validate_type('actual_volume_percent', actual_volume_percent, dict, Scalar)
        self._actual_volume_percent = self._get_object(Scalar, actual_volume_percent)

    @actual_volume_percent.deleter
    def actual_volume_percent(self):
        self._actual_volume_percent = None

    @property
    def actual_number_percent(self):
        return self._actual_number_percent

    @actual_number_percent.setter
    def actual_number_percent(self, actual_number_percent):
        self._validate_type('actual_number_percent', actual_number_percent, dict, Scalar)
        self._actual_number_percent = self._get_object(Scalar, actual_number_percent)

    @actual_number_percent.deleter
    def actual_number_percent(self):
        self._actual_number_percent = None

    @property
    def ideal_mass_percent(self):
        return self._ideal_mass_percent

    @ideal_mass_percent.setter
    def ideal_mass_percent(self, ideal_mass_percent):
        self._validate_type('ideal_mass_percent', ideal_mass_percent, dict, Scalar)
        self._ideal_mass_percent = self._get_object(Scalar, ideal_mass_percent)

    @ideal_mass_percent.deleter
    def ideal_mass_percent(self):
        self._ideal_mass_percent = None

    @property
    def ideal_volume_percent(self):
        return self._ideal_volume_percent

    @ideal_volume_percent.setter
    def ideal_volume_percent(self, ideal_volume_percent):
        self._validate_type('ideal_volume_percent', ideal_volume_percent, dict, Scalar)
        self._ideal_volume_percent = self._get_object(Scalar, ideal_volume_percent)

    @ideal_volume_percent.deleter
    def ideal_volume_percent(self):
        self._ideal_volume_percent = None

    @property
    def ideal_number_percent(self):
        return self._ideal_number_percent

    @ideal_number_percent.setter
    def ideal_number_percent(self, ideal_number_percent):
        self._validate_type('ideal_number_percent', ideal_number_percent, dict, Scalar)
        self._ideal_number_percent = self._get_object(Scalar, ideal_number_percent)

    @ideal_number_percent.deleter
    def ideal_number_percent(self):
        self._ideal_number_percent = None



import json


class PifEncoder(json.JSONEncoder):
    """
    Class to convert physical information objects to json.
    """

    def default(self, obj):
        """
        Convert an object to a form ready to dump to json.

        :param obj: Object being serialized. The type of this object must be one of the following: None; a single
        object derived from the Pio class; or a list of objects, each derived from the Pio class.
        :return: List of dictionaries, each representing a physical information object, ready to be serialized.
        """
        if obj is None:
            return []
        elif isinstance(obj, list):
            return [i.as_dictionary() for i in obj]
        else:
            return obj.as_dictionary()



import numbers
from six import string_types
from pypif.util.serializable import Serializable


class Pio(Serializable):
    """
    Base class for all physical information objects.
    """

    def __init__(self, tags=None, **kwargs):
        """
        Constructor. This is used to set an fields on a pio that do not have first level support.

        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of field names to values.
        """
        self._tags = None
        self.tags = tags
        for i in kwargs.keys():
            setattr(self, i, kwargs[i])

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        self._validate_list_type('tags', tags, string_types, numbers.Number)
        self._tags = tags

    @tags.deleter
    def tags(self):
        self._tags = None

    def _validate_type(self, name, obj, *args):
        """
        Helper function that checks the input object type against each in a list of classes. This function
        also allows the input value to be equal to None.

        :param name: Name of the object.
        :param obj: Object to check the type of.
        :param args: List of classes.
        :raises TypeError: if the input object is not of any of the allowed types.
        """
        if obj is None:
            return
        for arg in args:
            if isinstance(obj, arg):
                return
        raise TypeError(self.__class__.__name__ + '.' + name + ' is of type ' + type(obj).__name__ +
                        '. Must be equal to None or one of the following types: ' +
                        ', '.join([arg.__name__ for arg in args]))

    def _validate_type_not_null(self, name, obj, *args):
        """
        Helper function that checks the input object type against each in a list of classes.

        :param name: Name of the object.
        :param obj: Object to check the type of.
        :param args: List of classes.
        :raises TypeError: if the input object is not of any of the allowed types.
        """
        for arg in args:
            if isinstance(obj, arg):
                return
        raise TypeError(self.__class__.__name__ + '.' + name + ' is of type ' + type(obj).__name__ +
                        '. Must be one of the following types [' +
                        ', '.join([arg.__name__ for arg in args]) + ']')

    def _validate_list_type(self, name, obj, *args):
        """
        Helper function that checks the input object type against each in a list of classes, or if the input object
        is a list, each value that it contains against that list.

        :param name: Name of the object.
        :param obj: Object to check the type of.
        :param args: List of classes.
        :raises TypeError: if the input object is not of any of the allowed types.
        """
        if obj is None:
            return
        if isinstance(obj, list):
            for i in obj:
                self._validate_type_not_null(name,  i, *args)
        else:
            self._validate_type(name, obj, *args)

    def _validate_nested_list_type(self, name, obj, nested_level, *args):
        """
        Helper function that checks the input object as a list then recursively until nested_level is 1.

        :param name: Name of the object.
        :param obj: Object to check the type of.
        :param nested_level: Integer with the current nested level.
        :param args: List of classes.
        :raises TypeError: if the input object is not of any of the allowed types.
        """
        if nested_level <= 1:
            self._validate_list_type(name, obj, *args)
        else:
            if obj is None:
                return
            if not isinstance(obj, list):
                raise TypeError(self.__class__.__name__ + '.' + name + ' contains value of type ' +
                                type(obj).__name__ + ' where a list is expected')
            for sub_obj in obj:
                self._validate_nested_list_type(name, sub_obj, nested_level - 1, *args)



from six import string_types
from pypif.obj.common.pio import Pio


class DisplayItem(Pio):
    """
    Representation of a display item (table or figure) in a referenced work.
    """

    def __init__(self, number=None, title=None, caption=None, tags=None, **kwargs):
        """
        Constructor.

        :param number: String with the number of the item in the referenced work.
        :param title: Title of the item.
        :param caption: Caption associated with the item.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(DisplayItem, self).__init__(tags=tags, **kwargs)
        self._number = None
        self.number = number
        self._title = None
        self.title = title
        self._caption = None
        self.caption = caption

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, number):
        self._validate_type('number', number, string_types)
        self._number = number

    @number.deleter
    def number(self):
        self._number = None

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self._validate_type('title', title, string_types)
        self._title = title

    @title.deleter
    def title(self):
        self._title = None

    @property
    def caption(self):
        return self._caption

    @caption.setter
    def caption(self, caption):
        self._validate_type('caption', caption, string_types)
        self._caption = caption

    @caption.deleter
    def caption(self):
        self._caption = None



from pypif.obj.system.chemical.chemical_system import ChemicalSystem


class Alloy(ChemicalSystem):
    """
    Information about an alloy.
    """

    def __init__(self, uid=None, names=None, ids=None, source=None, quantity=None, chemical_formula=None,
                 composition=None, properties=None, preparation=None, sub_systems=None, phases=None,
                 references=None, contacts=None, licenses=None, tags=None, **kwargs):
        """
        Constructor.

        :param uid: String with the permanent ID for this record.
        :param names: List of strings with common names of the alloy.
        :param ids: List of dictionaries, strings, numbers, or :class:`.Id` objects that identify the alloy.
        :param source: Dictionary, string, or :class:`.Source` object with the source of the alloy.
        :param quantity: Dictionary or :class:`.Quantity` object with the quantity of the alloy.
        :param chemical_formula: String with the chemical formula.
        :param composition: List of dictionaries or :class:`.Composition` objects that describe the composition of
                the alloy.
        :param properties: List of dictionaries or :class:`.Property` objects with properties of the alloy.
        :param preparation: List of dictionaries or :class:`.ProcessStep` objects with the preparation
                information of the alloy.
        :param sub_systems: List of dictionaries or :class:`.System` objects with the subsystems of the alloy.
        :param phases: List of dictionaries or :class:`.AlloyPhase` objects with the phases of the alloy.
        :param references: List of dictionaries or :class:`.Reference` objects where information about the alloy
                is published.
        :param contacts: List of dictionaries, strings, or :class:`.Person` objects with people to contact for
                information about the alloy.
        :param licenses: List of dictionaries, strings, or :class:`.License` objects with licensing information
                for data about the alloy.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Alloy, self).__init__(uid=uid, names=names, ids=ids, source=source, quantity=quantity,
                                    chemical_formula=chemical_formula, composition=composition, properties=properties,
                                    preparation=preparation, sub_systems=sub_systems, references=references,
                                    contacts=contacts, licenses=licenses, tags=tags, **kwargs)
        self.phases = phases
        self.category = kwargs['category'] if 'category' in kwargs else 'system.chemical.alloy'

    @property
    def phases(self):
        return self._sub_systems

    @phases.setter
    def phases(self, phases):
        self.sub_systems = phases

    @phases.deleter
    def phases(self):
        del self.sub_systems



from pypif.util.case import to_camel_case
from pypif.util.case import keys_to_snake_case


class Serializable(object):
    """
    Base class for objects that can be serialized.
    """

    def as_dictionary(self):
        """
        Convert this object to a dictionary with formatting appropriate for a PIF.

        :returns: Dictionary with the content of this object formatted for a PIF.
        """
        return {to_camel_case(i): Serializable._convert_to_dictionary(self.__dict__[i])
                for i in self.__dict__ if self.__dict__[i] is not None}

    @staticmethod
    def _convert_to_dictionary(obj):
        """
        Convert obj to a dictionary with formatting appropriate for a PIF. This function attempts to treat obj as
        a Pio object and otherwise returns obj.

        :param obj: Object to convert to a dictionary.
        :returns: Input object as a dictionary or the original object.
        """
        if isinstance(obj, list):
            return [Serializable._convert_to_dictionary(i) for i in obj]
        elif hasattr(obj, 'as_dictionary'):
            return obj.as_dictionary()
        else:
            return obj

    @staticmethod
    def _get_object(class_, obj):
        """
        Helper function that returns an object, or if it is a dictionary, initializes it from class_.

        :param class_: Class to use to instantiate object.
        :param obj: Object to process.
        :return: One or more objects.
        """
        if isinstance(obj, list):
            return [Serializable._get_object(class_, i) for i in obj]
        elif isinstance(obj, dict):
            return class_(**keys_to_snake_case(obj))
        else:
            return obj



from six import string_types
from pypif.obj.common.pio import Pio
from pypif.obj.common.value import Value
from pypif.obj.common.software import Software
from pypif.obj.common.instrument import Instrument


class ProcessStep(Pio):
    """
    Information about a single step in a processing pipeline.
    """

    def __init__(self, name=None, details=None, instruments=None, software=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: String with the name of the process step.
        :param details: List of dictionaries or :class:`.Value` objects with the details of the step.
        :param instruments: List of dictionaries or :class:`.Instrument` objects.
        :param software: List of dictionaries or :class:`.Software` objects.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(ProcessStep, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._details = None
        self.details = details
        self._instruments = None
        self.instruments = instruments
        self._software = None
        self.software = software

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, string_types)
        self._name = name

    @name.deleter
    def name(self):
        self._name = None

    @property
    def details(self):
        return self._details

    @details.setter
    def details(self, details):
        self._validate_list_type('details', details, dict, Value)
        self._details = self._get_object(Value, details)

    @details.deleter
    def details(self):
        self._details = None

    @property
    def instruments(self):
        return self._instruments

    @instruments.setter
    def instruments(self, instruments):
        self._validate_list_type('instruments', instruments, dict, Instrument)
        self._instruments = self._get_object(Instrument, instruments)

    @instruments.deleter
    def instruments(self):
        self._instruments = None

    @property
    def software(self):
        return self._software

    @software.setter
    def software(self, software):
        self._validate_list_type('software', software, dict, Software)
        self._software = self._get_object(Software, software)



from six import string_types
from pypif.obj.common.name import Name
from pypif.obj.common.pio import Pio


class Person(Pio):
    """
    Information about a person.
    """

    def __init__(self, name=None, email=None, orcid=None, tags=None, **kwargs):
        """
        Constructor.

        :param name: :class: Dictionary or `.Name` object for the person.
        :param email: String with the email address of the person.
        :param orcid: String with the `ORCID <a href="http://orcid.org">ORCID</a>`_ identifier of the person.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Person, self).__init__(tags=tags, **kwargs)
        self._name = None
        self.name = name
        self._email = None
        self.email = email
        self._orcid = None
        self.orcid = orcid

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._validate_type('name', name, dict, Name)
        self._name = self._get_object(Name, name)

    @name.deleter
    def name(self):
        self._name = None

    @property
    def email(self):
        return self._email

    @email.setter
    def email(self, email):
        self._validate_type('email', email, string_types)
        self._email = email

    @email.deleter
    def email(self):
        self._email = None

    @property
    def orcid(self):
        return self._orcid

    @orcid.setter
    def orcid(self, orcid):
        self._validate_type('orcid', orcid, string_types)
        self._orcid = orcid

    @orcid.deleter
    def orcid(self):
        self._orcid = None



import re


_first_camel_case_regex = re.compile('(.)([A-Z][^A-Z]+)')
_second_camel_case_regex = re.compile('([^A-Z_])([A-Z])')


def to_camel_case(snake_case_string):
    """
    Convert a string from snake case to camel case. For example, "some_var" would become "someVar".

    :param snake_case_string: Snake-cased string to convert to camel case.
    :returns: Camel-cased version of snake_case_string.
    """
    parts = snake_case_string.lstrip('_').split('_')
    return parts[0] + ''.join([i.title() for i in parts[1:]])


def to_capitalized_camel_case(snake_case_string):
    """
    Convert a string from snake case to camel case with the first letter capitalized. For example, "some_var"
    would become "SomeVar".

    :param snake_case_string: Snake-cased string to convert to camel case.
    :returns: Camel-cased version of snake_case_string.
    """
    parts = snake_case_string.split('_')
    return ''.join([i.title() for i in parts])


def to_snake_case(camel_case_string):
    """
    Convert a string from camel case to snake case. From example, "someVar" would become "some_var".

    :param camel_case_string: Camel-cased string to convert to snake case.
    :return: Snake-cased version of camel_case_string.
    """
    first_pass = _first_camel_case_regex.sub(r'\1_\2', camel_case_string)
    return _second_camel_case_regex.sub(r'\1_\2', first_pass).lower()


def keys_to_snake_case(camel_case_dict):
    """
    Make a copy of a dictionary with all keys converted to snake case. This is just calls to_snake_case on
    each of the keys in the dictionary and returns a new dictionary.

    :param camel_case_dict: Dictionary with the keys to convert.
    :type camel_case_dict: Dictionary.

    :return: Dictionary with the keys converted to snake case.
    """
    return dict((to_snake_case(key), value) for (key, value) in camel_case_dict.items())



from pypif.obj.common.display_item import DisplayItem
from pypif.obj.common.id import Id
from pypif.obj.common.instrument import Instrument
from pypif.obj.common.license import License
from pypif.obj.common.method import Method
from pypif.obj.common.name import Name
from pypif.obj.common.pages import Pages
from pypif.obj.common.person import Person
from pypif.obj.common.process_step import ProcessStep
from pypif.obj.common.property import Property
from pypif.obj.common.quantity import Quantity
from pypif.obj.common.reference import Reference
from pypif.obj.common.scalar import Scalar
from pypif.obj.common.software import Software
from pypif.obj.common.source import Source
from pypif.obj.common.value import Value



import numbers
from six import string_types
from pypif.obj.common.pio import Pio
from pypif.obj.common.scalar import Scalar


class Composition(Pio):
    """
    Class to store information about an element in a composition vector using weight percents or atomic percents.
    """

    def __init__(self, element=None, actual_weight_percent=None, actual_atomic_percent=None, ideal_weight_percent=None,
                 ideal_atomic_percent=None, tags=None, **kwargs):
        """
        Constructor.

        :param element: Symbol of the element.
        :param actual_weight_percent: Actual percentage of the weight of the chemical that is this element
                - dictionary, string, number, or :class:`.Scalar` object.
        :param actual_atomic_percent: Actual percentage of the atoms in the chemical that are this element
                - dictionary, string, number, or :class:`.Scalar` object.
        :param ideal_weight_percent: Ideal percentage of the weight of the chemical that is this element
                - dictionary, string, number, or :class:`.Scalar` object.
        :param ideal_atomic_percent: Ideal percentage of the atoms in the chemical that are this element
                - dictionary, string, number, or :class:`.Scalar` object.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Composition, self).__init__(tags=tags, **kwargs)
        self._element = None
        self.element = element
        self._actual_weight_percent = None
        self.actual_weight_percent = actual_weight_percent
        self._actual_atomic_percent = None
        self.actual_atomic_percent = actual_atomic_percent
        self._ideal_weight_percent = None
        self.ideal_weight_percent = ideal_weight_percent
        self._ideal_atomic_percent = None
        self.ideal_atomic_percent = ideal_atomic_percent

    @property
    def element(self):
        return self._element

    @element.setter
    def element(self, element):
        self._validate_type('element', element, string_types)
        self._element = element

    @element.deleter
    def element(self):
        self._element = None
        
    @property
    def actual_weight_percent(self):
        return self._actual_weight_percent
        
    @actual_weight_percent.setter
    def actual_weight_percent(self, actual_weight_percent):
        self._validate_type('actual_weight_percent', actual_weight_percent, dict, string_types,
                            numbers.Number, Scalar)
        self._actual_weight_percent = self._get_object(Scalar, actual_weight_percent)
    
    @actual_weight_percent.deleter
    def actual_weight_percent(self):
        self._actual_weight_percent = None
    
    @property
    def actual_atomic_percent(self):
        return self._actual_atomic_percent
        
    @actual_atomic_percent.setter
    def actual_atomic_percent(self, actual_atomic_percent):
        self._validate_type('actual_atomic_percent', actual_atomic_percent, dict, string_types,
                            numbers.Number, Scalar)
        self._actual_atomic_percent = self._get_object(Scalar, actual_atomic_percent)
    
    @actual_atomic_percent.deleter
    def actual_atomic_percent(self):
        self._actual_atomic_percent = None
    
    @property
    def ideal_weight_percent(self):
        return self._ideal_weight_percent
        
    @ideal_weight_percent.setter
    def ideal_weight_percent(self, ideal_weight_percent):
        self._validate_type('ideal_weight_percent', ideal_weight_percent, dict, string_types,
                            numbers.Number, Scalar)
        self._ideal_weight_percent = self._get_object(Scalar, ideal_weight_percent)
    
    @ideal_weight_percent.deleter
    def ideal_weight_percent(self):
        self._ideal_weight_percent = None
    
    @property
    def ideal_atomic_percent(self):
        return self._ideal_atomic_percent
        
    @ideal_atomic_percent.setter
    def ideal_atomic_percent(self, ideal_atomic_percent):
        self._validate_type('ideal_atomic_percent', ideal_atomic_percent, dict, string_types,
                            numbers.Number, Scalar)
        self._ideal_atomic_percent = self._get_object(Scalar, ideal_atomic_percent)
    
    @ideal_atomic_percent.deleter
    def ideal_atomic_percent(self):
        self._ideal_atomic_percent = None



from six import string_types
from pypif.obj.common.pio import Pio


class Source(Pio):
    """
    Information about the source of a system.
    """

    def __init__(self, producer=None, url=None, tags=None, **kwargs):
        """
        Constructor.

        :param producer: String with the name of the producer.
        :param url: String with the URL to the source of the system.
        :param tags: List of strings or numbers that are tags for this object.
        :param kwargs: Dictionary of fields that are not supported.
        """
        super(Source, self).__init__(tags=tags, **kwargs)
        self._producer = None
        self.producer = producer
        self._url = None
        self.url = url

    @property
    def producer(self):
        return self._producer

    @producer.setter
    def producer(self, producer):
        self._validate_type('producer', producer, string_types)
        self._producer = producer

    @producer.deleter
    def producer(self):
        self._producer = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._validate_type('url', url, string_types)
        self._url = url

    @url.deleter
    def url(self):
        self._url = None
